/* ************************************************************** *
 * Library for the spu, manages sending and receiving of messages *
 * from the ppu as well as reading/writing to memory in addition  *
 * to manage the entire neststep-runtime, the NestStep_init() and *
 * the NestStep_finalize() are in the ppu-part, the functions     * 
 * NestStep_spu_init() and NestStep_spu_finalize() only handles   *
 * getting and setting of the DataManager object so that it is    *
 * consistent with what is inside the ppu so variables can be     *
 * allocated and maintained over spu-program boundaries.          *
 * ************************************************************** */
#ifndef NESTSTEP_SPU_H
#define NESTSTEP_SPU_H

//spu specific functions and variables
#include "base_spu.h"
#include "combine_spu.h"
#include "import_export_spu.h"
#include "tree_builders_spu.h"


/* ************************************************************** *
 * NestStep functions                                             *
 * ************************************************************** */

//NestStep init initializes the datastructures so userspace programs
//can access created variables and Neststep finalize clears up allocated memory
void NestStep_SPU_init(Addr64 argp);
void NestStep_SPU_finalize(void);

// NestStep begin step, it is only allowed to update shared variables inside
// a step-statement, this also sets the scope of the variables declared
// this is stored inside thisgroup->path
void NestStep_step(); 
// NestStep end step, reduces the thisgroup->path so that it is on the
// correct level.
void NestStep_end_step(); 

//retreives the rank of the SPU in the current group
int NestStep_get_rank(void);

//retreives the group size of the current group
int NestStep_get_size(void);

//get the group number
int NestStep_get_group_id(void);

//return a pointer to the parent group
Group* NestStep_get_parent_group(void);

void NestStep_get_all_ls(Ls_pointers *ls);

//NestStep_neststep(int k) is a nested step that creates k subgroups
//void NestStep_neststep(int k) //NOTE: may be implemented later, or not.

//NestStep_neststep_w(int k, float weight[]) is a nested step that 
//creates k subgroups with groupsizes that is specified by the weight array
//NestStep_neststep_w(int k, float weight[]); //NOTE: may be implemented later, no.

//NestStep_neststep_c(int k, int color) is a nested step where color 
//decides which subgroup the processor ends up in
//NestStep_neststep_c(int k, int color); //NOTE: may be implemented later, no.


//functions to load the variables used in earlier spu-programs
//these functions assume that the dataManager has just been imported
//from the ppu and that it has all it's pointers in global-memory space.
//it will rewrite the loaded dataManager to point to local variables
//and load the variables from main memory
void import_shVar();
void import_pVar();
void import_pArray();
void import_bArray();
void import_cArray();
void import_repArray();

/* ************************************************************** *
 * ShVar, Pvar, PArr, CArr, BArr and RArr Constructors and        *
 * Destructors, these will send a request to the PPU for          *
 * allocating a block of memory                                   *
 * ************************************************************** */

//create a new private variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace, _no
//is the variable number in case of duplicates (i.e. arrays)
PVar* new_PVar(Name _name, 
               int _type, 
               Addr64 _base);
void free_PVar(PVar* var);

//create a new shared variable, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace, _no
//is the variable number in case of duplicates (i.e. arrays)
ShVar* new_ShVar(Name _name, 
                 int _type, 
                 Addr64 _base);
void free_ShVar(ShVar* var);

//create a new private array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
PrivateArray* new_PrivateArray(Name _name, 
                               int _type, 
                               int _size, 
                               Addr64 _base);
void free_PrivateArray(PrivateArray* var);

//create a new replicated array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, base is the address of the "real" variable
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
RepArray* new_RepArray(Name _name, 
                       int _type, 
                       int _size, 
                       Addr64 _base);
void free_RepArray(RepArray* var);

//create a new block array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
BlockDistArray* new_BlockDistArray(Name _name, 
                                   int _type, 
                                   int _size);
void free_BlockDistArray(BlockDistArray* var);

//create a new cyclic array, name is the path to where it was created
//type is IVAR, FVAR or DVAR, size is the total size of the array, blocksize
//is the size of each segment that is chosen modulo the number of threads
//allocated, this will create a new variable on the heap so it
//doesn't need to take a variable, this is due to segmentation of the array
//and uname is the name by which it will be accessible from userspace
//size is the size of the array, _no
//is the variable number in case of duplicates (i.e. arrays)
CyclicDistArray* new_CyclicDistArray(Name _name, 
                                     int _type, 
                                     int _size, 
                                     int _blocksize);
void free_CyclicDistArray(CyclicDistArray* var);

/* ************************************************************** *
 * ShVar, PVar, PArr, RArr, CArr, BArr getters, setters           *
 * and barriers. The getters and setters have separate functions  *
 * for storing/retreiving integers, floats and doubles. these are *
 * named get_***I, store_***I and so on with F and D as suffixes  *
 * respectively.                                                  *
 * ************************************************************** */

//get/set a shared variable from/to the DataManager
void get_ShVarI(ShVar* _shvar, Addr64 lsaddr);
void get_ShVarF(ShVar* _shvar, Addr64 lsaddr);
void get_ShVarD(ShVar* _shvar, Addr64 lsaddr);
void store_ShVarI(ShVar* _shvar,Addr64 lsaddr);
void store_ShVarF(ShVar* _shvar,Addr64 lsaddr);
void store_ShVarD(ShVar* _shvar,Addr64 lsaddr);


//get/set a private variable from/to the DataManager
void get_PVarI(PVar* _pvar, Addr64 lsaddr);
void get_PVarF(PVar* _pvar, Addr64 lsaddr);
void get_PVarD(PVar* _pvar, Addr64 lsaddr);
void store_PVarI(PVar* _pvar,Addr64 lsaddr);
void store_PVarF(PVar* _pvar,Addr64 lsaddr);
void store_PVarD(PVar* _pvar,Addr64 lsaddr);


//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from a private array with the userspace-name name from/to the DataManager
void get_PArrI(PrivateArray* _parr,Addr64 lsaddr, int lbound, int ubound);
void get_PArrF(PrivateArray* _parr,Addr64 lsaddr, int lbound, int ubound);
void get_PArrD(PrivateArray* _parr,Addr64 lsaddr, int lbound, int ubound);
void store_PArrI(PrivateArray* _parr, Addr64 lsaddr,int lbound,int ubound);
void store_PArrF(PrivateArray* _parr, Addr64 lsaddr,int lbound,int ubound);
void store_PArrD(PrivateArray* _parr, Addr64 lsaddr,int lbound,int ubound);


//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from the segment that this node owns from a cyclic array with the 
//userspace-name name from/to the DataManager
void get_CArrI(CyclicDistArray* _carr,Addr64 lsaddr, int lbound, int ubound);
void get_CArrF(CyclicDistArray* _carr,Addr64 lsaddr, int lbound, int ubound);
void get_CArrD(CyclicDistArray* _carr,Addr64 lsaddr, int lbound, int ubound);
void store_CArrI(CyclicDistArray* _carr, Addr64 lsaddr,int lbound,int ubound);
void store_CArrF(CyclicDistArray* _carr, Addr64 lsaddr,int lbound,int ubound);
void store_CArrD(CyclicDistArray* _carr, Addr64 lsaddr,int lbound,int ubound);

//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from the segment that this node owns from a block array with the 
//userspace-name name from/to the DataManager
void get_BArrI(BlockDistArray* _barr,Addr64 lsaddr, int lbound, int ubound);
void get_BArrF(BlockDistArray* _barr,Addr64 lsaddr, int lbound, int ubound);
void get_BArrD(BlockDistArray* _barr,Addr64 lsaddr, int lbound, int ubound);
void store_BArrI(BlockDistArray* _barr, Addr64 lsaddr,int lbound,int ubound);
void store_BArrF(BlockDistArray* _barr, Addr64 lsaddr,int lbound,int ubound);
void store_BArrD(BlockDistArray* _barr, Addr64 lsaddr,int lbound,int ubound);


//get/set data between the indexes lbound(lower bound) to ubound(upper bound)
//from a replicated array with the userspace-name name from/to the DataManager
void get_RArrI(RepArray* _rarr,Addr64 lsaddr, int lbound, int ubound);
void get_RArrF(RepArray* _rarr,Addr64 lsaddr, int lbound, int ubound);
void get_RArrD(RepArray* _rarr,Addr64 lsaddr, int lbound, int ubound);
void store_RArrI(RepArray* _rarr, Addr64 lsaddr,int lbound,int ubound);
void store_RArrF(RepArray* _rarr, Addr64 lsaddr,int lbound,int ubound);
void store_RArrD(RepArray* _rarr, Addr64 lsaddr,int lbound,int ubound);

//Barriers for managing DMA transfers to/from the main memory
void wait_PVar_complete(PVar* _pvar,Addr64 lsaddr);
void wait_ShVar_complete(ShVar* _shvar,Addr64 lsaddr);
void wait_BArr_complete(BlockDistArray* _barr,Addr64 lsaddr);
void wait_CArr_complete(CyclicDistArray* _carr,Addr64 lsaddr);
void wait_PArr_complete(PrivateArray* _parr,Addr64 lsaddr);
void wait_RArr_complete(RepArray* _rarr,Addr64 lsaddr);

//functions for managing arrays
//these functions set all elements in the arrays to zero.
void set_RepArray_zero(RepArray *array);
void set_BlockDistArray_zero(BlockDistArray *array);
void set_CyclicDistArray_zero(CyclicDistArray *array);

//TODO: read_request_b, read_request_c
void read_request_b(PrivateArray* _parr, BlockDistArray* _barr, int lbound, int ubound);
void read_request_c(PrivateArray* _parr, CyclicDistArray* _carr, int lbound, int ubound);
//TODO: write_request_b, write_request_c
void write_request_b(PrivateArray* _parr, BlockDistArray* _barr, int lbound, int ubound);
void write_request_c(PrivateArray* _parr, CyclicDistArray* _carr, int lbound, int ubound);

//these functions transform a local index to a global index and
//from global to local respectively.
int global_local_index_b(BlockDistArray *array, int index);
int local_global_index_b(BlockDistArray *array, int index);
int global_local_index_c(CyclicDistArray *array, int index);
int local_global_index_c(CyclicDistArray *array, int index);


#endif //NESTSTEP_SPU_H
