/* ********************************************** * 
 * Jacobi test-program, this will LP-filter a     * 
 * sampled signal in the signal-domain, i.e.      *
 * multiply it with a sampled sinc -1 4 4 -1      *
 * ********************************************** */

#include <stdio.h>
#include <../../NestStep_spulib/neststep_spu.h>

//#include<time.h>
//#include<sys/time.h>

int MY_RANK;
int GROUP_SIZE;

//#define N 1966080
#define N 1966080*4

// Used for transforming clock-ticks to
// seconds, gotten from /proc/cpuinfo
#define TIMEBASE 14318000
// Initializing value for the counter
#define DEC_INIT_VAL 0XFFFFFFFF

int ublock1, ublock2, ublocktot;
int twait1, twait2, twaittot;
int ncomb1, ncomb2, ncombtot;
//sh float a[N]</>;  // blockwise distributed shared array, resident in main memory
//sh float b[N]</>;  // blockwise distributed shared array, resident in main memory

BlockDistArray* a, *aorig;
BlockDistArray* b, *borig;

// implicit assumption: each local array partition has at least 4 
// and at least 4*blocksize elements, blocksize at least 5.
// i.e., N >= 4*#.

//#define blocksize 1024   // unit of size for DMA block transfers, tuning parameter
//#define blocksize 2048   // unit of size for DMA block transfers, tuning parameter
#define blocksize 4096   // unit of size for DMA block transfers, tuning parameter

// We need on a SPE 6 buffers of size blocksize floats, plus O(1) space
// for other data.

#define EPS 0.01
#define left_boundary_value 0.0
#define right_boundary_value 1.0

// all the following variables are SPE-local (and static)
float left[4], right[4];
PrivateArray *pleft, *pright;
   // private mirrors for nearest 2 left resp. right neighbor elements
   // order:  ... left[0], left[1], local elements, right[0], right[1], ...
//sh float maxdiff;
float maxdiff;
ShVar* _maxdiff;
float la[4][blocksize];  // 4 cyclically managed buffers for a-blocks:
                           //    last, current, next, nextnext(being reloaded).
               // each buffer is of size blocksize
int la_currentblock = 0;   // index in 0..3 for current position of a
int la_nextblock, la_lastblock, la_nextnextblock;
float lb[2][blocksize];  // 2 buffers for b-blocks, each buffer is of size blocksize
int lb_currentblock = 0;   // index in 0..1
float divider = 0.16666667;
/**
 * Jacobi-update of a block A of bsize floats, result block written to B.
 * Left and right boundary elements delivered by L and R.
 * The maximum occurring difference in A vs B values is updated in mdif.
 */
float max(float a, float b){
    return (a>b)?a:b;
}
void update_block( float *B, float *A, float *L, float *R, int bsize, float *mdif )
{
    int i;
    B[0] = (-L[bsize-2] + 4.0*L[bsize-1] + 4.0*A[1] - A[2]) * divider;
    B[1] = (-L[bsize-1] + 4.0*A[0] + 4.0*A[2] - A[3]) * divider;
    *mdif = max( fabs(B[0]-A[0]), fabs(B[1]-A[1]) );
    for (i=2; i<bsize-2; i++) {
	B[i] = (-A[i-2] + 4.0*A[i-1] + 4.0*A[i+1] - A[i+2]) * divider;
	*mdif = max( *mdif, fabs( B[i] - A[i] ) );
    }
    B[bsize-2] = (-A[bsize-4] + 4.0 * A[bsize-3] + 4.0*A[bsize-1] - R[0]) * divider;
    B[bsize-1] = (-A[bsize-3] + 4.0 * A[bsize-2] + 4.0*R[0] - R[1]) * divider;
    *mdif = max( *mdif, fabs( B[bsize-2]-A[bsize-2] ) );
    *mdif = max( *mdif, fabs( B[bsize-1]-A[bsize-1] ) );
}
                          
/**
 * Jacobi 1d four-stencil computation
 * blocked for limited SPE size with explicit DMA transfers 
 * for manual prefetching.
 */
int localsize(BlockDistArray* arr){
    return arr->localsize;
}
int localoffset(BlockDistArray* arr){
    return NestStep_get_rank()* arr->localsize;
}
void jacobi( void )
{
    int lap = 0;

    int mysize = localsize(a);     // size of my local partition of a
    // should evaluate to  N / # + ($<N%#)? 1:0;
    int myfirst = localoffset(a);     // $ * mysize if all have same partition size
    int mylast = myfirst + mysize - 1;
    int ii;

    NestStep_step(); 
    {
	if (MY_RANK != 0){    // I have not the leftmost array partition
	    read_request_b(pleft,a,myfirst-2,myfirst-1);
	}
	else left[0] = left[1] = 0.0; // left_boundary_value;
	
	if (MY_RANK != GROUP_SIZE-1){  // I have not the rightmost array partition
	    read_request_b(pright,a,mylast+1,mylast+2);
	}
	else right[0] = right[1] = right_boundary_value;
    }
    ncomb1 = spu_read_decrementer();
    NestStep_combine(NULL,NULL); // mirror commits
    ncomb2 = spu_read_decrementer();
    ncombtot += ncomb1-ncomb2;
    NestStep_end_step();
    do {
	lap++;

	if(MY_RANK != 0){
	    get_PArrF(pleft,(Addr64)left,0,3); //only the two first
	                                       //elements are used,
	                                       //the other two are
	                                       //just for getting a
	                                       //16byte transfer
	    twait1 = spu_read_decrementer();
	    wait_PArr_complete(pleft,(Addr64)left);
	    twait2 = spu_read_decrementer();
	    twaittot+= twait1-twait2;
	}
	if(MY_RANK != GROUP_SIZE-1){
	    get_PArrF(pright,(Addr64)right,0,3); //only the two first
						 //elements are used,
						 //the other two are
						 //just for getting a
						 //16byte transfer
	    twait1 = spu_read_decrementer();
	    wait_PArr_complete(pright,(Addr64)right);
	    twait2 = spu_read_decrementer();
	    twaittot+= twait1-twait2;
	}
	
	NestStep_step(); 
	{
	    la_currentblock = 0;
	    la_nextblock = (la_currentblock + 1) % 4;
	    la_nextnextblock = (la_nextblock + 1) % 4;
	    lb_currentblock = 0;
	    get_BArrF(a,(Addr64)la[la_currentblock],
		      myfirst,myfirst+blocksize-1 );  // la0 is local buffer for A
	    get_BArrF(a,(Addr64)la[la_nextblock],
		      myfirst+blocksize,myfirst+2*blocksize-1 );
	    get_BArrF(a,(Addr64)la[la_nextnextblock],
		      myfirst+2*blocksize,myfirst+3*blocksize-1 );
	    maxdiff = 0;
	    
	    twait1= spu_read_decrementer();
	    wait_BArr_complete(a,(Addr64) la[la_currentblock] );
	    twait2 = spu_read_decrementer();
	    twaittot+=twait1-twait2;

	    // first 2 elements: 
	    ublock1 = spu_read_decrementer();
	    lb[lb_currentblock][0] = (-left[0] + 4.0*left[1] + 4.0*la[la_currentblock][1] - la[la_currentblock][2]) * divider;
	    lb[lb_currentblock][1] = (-left[1] + 4.0*la[la_currentblock][0] + 4.0*la[la_currentblock][2] - la[la_currentblock][3]) * divider;

	    maxdiff = max( fabs(lb[lb_currentblock][0]-la[la_currentblock][0]), 
			   fabs( lb[lb_currentblock][1]-la[la_currentblock][1] ) );
	    ublock2 = spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;

	    twait1 = spu_read_decrementer();
	    wait_BArr_complete(a,(Addr64)la[1]);
	    twait2 = spu_read_decrementer();
	    twaittot +=twait1-twait2;
        // first block, without first 2 elements:
	    ublock1 = spu_read_decrementer();
	    update_block( lb[lb_currentblock]+2,         // result block
			  la[la_currentblock]+2, // current block
			  
			  la[la_currentblock] - (blocksize - 2), // fictive predecessor block
			  la[la_nextblock], // successor block
			  blocksize-2, 
			  &maxdiff );

	    ublock2 = spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;	 
	    store_BArrF(b,(Addr64)lb[lb_currentblock], myfirst, myfirst+blocksize -1);
	    twait1 = spu_read_decrementer();
	    wait_BArr_complete(a,(Addr64)la[2]);   // prefetch
	    // blocks 2, ..., #blocks-1:
	    wait_BArr_complete(b,(Addr64)lb[0]);
	    twait2 = spu_read_decrementer();
	    twaittot+= twait1-twait2;

	    for (ii = myfirst + blocksize, la_currentblock = 1, lb_currentblock = 1;
		 ii < mylast - 1 - 2*blocksize;
		 ii+=blocksize, la_currentblock = (la_currentblock+1) % 4, 
		     lb_currentblock = 1 - lb_currentblock ) { 

		la_lastblock = (la_currentblock + 3) % 4;
		la_nextblock = (la_currentblock + 1) % 4;
		la_nextnextblock = (la_currentblock + 2) % 4;

		get_BArrF(a,(Addr64)la[la_nextnextblock],ii+2*blocksize,ii+3*blocksize-1);
		twait1 = spu_read_decrementer();
		wait_BArr_complete(a,(Addr64)la[la_nextnextblock]);
		wait_BArr_complete(b,(Addr64)lb[lb_currentblock]);
		twait2 = spu_read_decrementer();
		twaittot+= twait1-twait2;
		ublock1 = spu_read_decrementer();
		update_block( (lb[lb_currentblock]), // result block
			      (la[la_currentblock]), // source main block
			      (la[la_lastblock]), // predecessor block of a
			      (la[la_nextblock]), // successor block of a
			      blocksize,   // process full blocks here
			      &maxdiff );
		ublock2 = spu_read_decrementer();
		ublocktot+=ublock1-ublock2;

		store_BArrF(b,(Addr64)lb[lb_currentblock], 
			    ii,ii+ blocksize -1 );

	    // special treatment for next-to-last block?
	    la_lastblock = (la_currentblock+3)%4;
	    la_nextblock = (la_currentblock+1)%4;

	    ublock1 = spu_read_decrementer();
	    update_block( lb[lb_currentblock], // result block
			  la[la_currentblock], // source main block
			  la[la_lastblock], // predecessor block of a
			  la[la_nextblock], // successor block of a
			  blocksize,
			  &maxdiff );
	    ublock2 = spu_read_decrementer();
	    ublocktot+=ublock1-ublock2;

	    store_BArrF(b,(Addr64)lb[lb_currentblock], ii,ii+ blocksize -1 );
	    ii+=blocksize;

	    lb_currentblock = 1 - lb_currentblock;
	    twait1 = spu_read_decrementer();
	    wait_BArr_complete(b,(Addr64)lb[lb_currentblock]);
	    twait2 = spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    // last block, remaining mysize-ii <= blocksize owned elements:
	    // (assumption: remaining block has no less than 4 elements)
	    // la_currentblock, lb_currentblock was already incremented above:

	    la_lastblock = (la_lastblock+1)%4;
	    la_currentblock = (la_currentblock+1)%4;

	    ublock1 = spu_read_decrementer();
	    update_block( lb[lb_currentblock], // result block
			  la[la_currentblock], // source main block
			  la[la_lastblock], // predecessor block of a
			  right, // successor block of a
			  myfirst+mysize - ii,
			  &maxdiff );
	    ublock2 = spu_read_decrementer();
	    ublocktot+= ublock1-ublock2;

	    store_BArrF(b,(Addr64)lb[lb_currentblock], ii,ii+ blocksize -1);
	    lb_currentblock = 1 - lb_currentblock;
	    twait1=spu_read_decrementer();
	    wait_BArr_complete(b,(Addr64)lb[lb_currentblock]);
	    lb_currentblock = 1 - lb_currentblock;
	    wait_BArr_complete(b,(Addr64)lb[lb_currentblock]);
	    twait2=spu_read_decrementer();
	    twaittot+=twait1-twait2;
	    // Copy-back loop:

	    BlockDistArray* tmp = a;
	    a = b;
	    b = tmp;
	    if (MY_RANK != 0)    // I have not the leftmost array partition
		read_request_b(pleft,a,myfirst-2,myfirst-1);
	    
	    if (MY_RANK != GROUP_SIZE-1)  // I have not the rightmost array partition
		read_request_b(pright,a,mylast+1,mylast+2);

	    store_ShVarF(_maxdiff,(Addr64)&maxdiff);
	    //	    printf("maxdiff before: %f\n",maxdiff);
	    NestStep_ShVar_attach(_maxdiff, COMBINE_MAX, NULL, NO_PREFIX,NULL, NOT_UPDATED, NULL);
	    twait1 = spu_read_decrementer();
	    wait_ShVar_complete(_maxdiff,(Addr64)&maxdiff);
	    twait2 = spu_read_decrementer();
	    twaittot+=twait1-twait2;
	}
	ncomb1 = spu_read_decrementer();
	NestStep_combine(NULL,NULL);
	ncomb2 = spu_read_decrementer();
	ncombtot+=ncomb1-ncomb2;

	NestStep_end_step();

	get_ShVarF(_maxdiff,(Addr64)&maxdiff);
	twait1 = spu_read_decrementer();
	wait_ShVar_complete(_maxdiff,(Addr64)&maxdiff);
	twait2 = spu_read_decrementer();
	twaittot+=twait1-twait2;

	if(lap == 24){
	    break;
	}

  } while ( maxdiff > EPS || TRUE);

} // jacobi


void read_file(FILE* file, BlockDistArray* barr, int lbound, int ubound){
    int i;
    int size = ubound-lbound+1;
    int noblocks = size/blocksize;
    float* buffer = malloc(blocksize*sizeof(float));

    fseek(file,lbound*sizeof(float),SEEK_SET);

    for(i=0;i<noblocks;i++){
	fread(buffer,sizeof(float),blocksize,file);
	store_BArrF(barr,(Addr64)buffer,lbound+i*blocksize,lbound+(i+1)*blocksize-1);
	wait_BArr_complete(a,(Addr64)buffer);
    }
    free(buffer);
}
void write_file(FILE* file, BlockDistArray* barr, int lbound, int ubound){
    int i;
    int size = ubound-lbound+1;
    int noblocks = size/blocksize;
    float* buffer = malloc(blocksize*sizeof(float));
    for(i=0;i<noblocks;i++){
	get_BArrF(barr,(Addr64)buffer,lbound+i*blocksize,lbound+(i+1)*blocksize-1);
	wait_BArr_complete(a,(Addr64)buffer);
	fwrite(buffer,sizeof(float),blocksize,file);
    }
    free(buffer);
}

int main(unsigned long long speid, Addr64 argp, Addr64 envp)
{
    int tmp_time1, tmp_time2, tot_time;
    ublock1 = ublock2 = ublocktot = 0;
    twait1 = twait2 = twaittot = 0;
    ncomb1 = ncomb2 = ncombtot = 0;

    int i;
    NestStep_SPU_init(argp);
    Name name;
    name.procedure = 1;
    name.relative = 0;

    spu_write_decrementer(DEC_INIT_VAL);

    //defined at top of file to make code more readable
    MY_RANK = NestStep_get_rank();
    GROUP_SIZE = NestStep_get_size();

    //create variables
    a = new_BlockDistArray(name,FVAR,N);
    aorig = a;
    name.relative++;
    b = new_BlockDistArray(name,FVAR,N);
    borig = b;
    name.relative++;
    pleft = new_PrivateArray(name,FVAR,4,(Addr64)(void*)NULL);
    name.relative++;
    pright = new_PrivateArray(name,FVAR,4,(Addr64)(void*)NULL);
    name.relative++;
    _maxdiff = new_ShVar(name,FVAR,(Addr64)(void*)NULL);
    name.relative++;

    char itmp[21];

    sprintf(itmp,"a.in");
    char otmp[22];
    sprintf(otmp,"%i.out",NestStep_get_rank());
    tot_time = 0;
    for(i=0;i<3;i++){

	FILE * ifile = fopen(itmp, "r");
	if(!ifile){
	    printf("could not open file\n");
	    exit(1);
	}

	read_file(ifile,a,localoffset(a),localsize(a)+localoffset(a));
	fclose(ifile);

	tmp_time1 = spu_read_decrementer();

	jacobi();
	tmp_time2 = spu_read_decrementer();
	tot_time += tmp_time1-tmp_time2; 

	FILE * ofile = fopen(otmp, "w+");
	if(!ofile){
	    printf("could not open file\n");
	    exit(1);
	}

	write_file(ofile,a,localoffset(a),localsize(a)+localoffset(a));
	fclose(ofile);
    }
    if(MY_RANK == 0){
	printf("Tot executiontime: %u, %f\n",tot_time, 
	       ((float)tot_time)/TIMEBASE);
	printf("Time spent calculating: %u, %f\n",ublocktot,
	       ((float)ublocktot)/TIMEBASE);
	printf("Time spent waiting: %u, %f\n",twaittot,
	       ((float)twaittot)/TIMEBASE);
	printf("Time spent in combine: %u, %f\n",ncombtot,
	       ((float)ncombtot)/TIMEBASE);
    }
    NestStep_SPU_finalize();
}

